#include <stdio.h>
#ifdef READLINE_INSTALLED
#include <readline/readline.h>
#include <readline/history.h>
#else
#include "readline.c"
#endif

#define DEFAULT_SORT_STYLE 0
#define OLD_SORT_STYLE 1

/* command line */
char *line;
char *readline (char *prompt);

/* on-line sorted dictionary */
unsigned char map[256][4];
unsigned char **list, **clist;
int nword, assigned;
int sortstyle;

/* function prototypes */
void initialize ();
void createMap (int style);
void createSwarMap (int style);
void createBanjonMap (int style);
void createJukMap (int style);
void createPuncMap (int style);
void trim (char *line);
void process (char *line);
void exitBwsort (FILE *fp);
void printVer (FILE *fp);
void printHelp (FILE *fp, char topic[]);
void convert (unsigned char inp[], unsigned char *outp);
int compare (unsigned char one[], unsigned char two[], int flag);
void insert (unsigned char *nextline, unsigned char *convline);
void loadFile (char fname[]);
void addFile (char fname[]);
void showList ();
void saveFile (char fname[]);

void initialize() {
   char *envstyle = NULL;

   list = clist = NULL;
   nword = assigned = 0;
   envstyle = (char *)getenv("BWSORTSTYLE");
   if (envstyle == NULL) sortstyle = DEFAULT_SORT_STYLE;
   else if (!strcmp(envstyle, "DEFAULT")) sortstyle = DEFAULT_SORT_STYLE;
   else if (!strcmp(envstyle, "OLD")) sortstyle = OLD_SORT_STYLE;
   else {
      fprintf(stderr,"The value \"%s\" of the environment variable BWSORTSTYLE\n", envstyle);
      fprintf(stderr,"is not understood. Using DEFAULT ...\n");
      sortstyle = DEFAULT_SORT_STYLE;
   }
}

void createMap (int style) {
   int i;

   for (i=0; i<256; i++) {
      map[i][0] = i;
      map[i][1] = map[i][2] = map[i][3] = 0;
   }
   createSwarMap(style);
   createBanjonMap(style);
   createJukMap(style);
   createPuncMap(style);
}

void createSwarMap (int style) {
   int i;

   /* This procedure is independent of style */

   i = 64;
   map['_'][0] = i++;                /* hasanta */
   map['a'][0] = i++;                /* a */
   map[9][0] = i++;                  /* ae */
   map['A'][0] = i++;                /* aa */
   map[1][0] = map['i'][0] = i++;    /* hraswa i */
   map[2][0] = map['I'][0] = i++;    /* dirgha i */
   map[3][0] = map['u'][0] = i++;    /* hraswa u */
   map[4][0] = map['U'][0] = i++;    /* dirgha u */
   map[5][0] = map['W'][0] = i++;    /* ri */
   map[6][0] = map['e'][0] = i++;    /* e */
   map[10][0] = i++;                 /* alternative spelling of ae */
   map[7][0] = map['E'][0] = i++;    /* oi */
   map[8][0] = i++;                  /* o */
   map['o'][0] = map['O'][0] = i++;  /* ou */
}

void createBanjonMap (int style) {
   int i;

   i = 78;
   if (style == OLD_SORT_STYLE) {
      map['M'][0] = i++;   /* anuswar */
      map['H'][0] = i++;   /* bisarga */
      map['w'][0] = i++;   /* chandrabindu */
   }

   /* k-barga */
   map['k'][0] = i++; map['K'][0] = i++;
   map['g'][0] = i++; map['G'][0] = i++;
   map['q'][0] = i++;

   /* cha-barga */
   map['c'][0] = i++; map['C'][0] = i++;
   map['j'][0] = i++; map['J'][0] = i++;
   map['Q'][0] = i++;

   if (style == OLD_SORT_STYLE) i = 97;

   /* Ta-barga */
   map['T'][0] = i++; map['Z'][0] = i++;
   map['D'][0] = i++;
   if (style == OLD_SORT_STYLE) map['R'][0] = i++; /* Da-e shunya ra is identified with Da */
   if (style == DEFAULT_SORT_STYLE) i = 97;
   map['X'][0] = i++;
   if (style == OLD_SORT_STYLE) map['V'][0] = i++; /* Dha-e shunya ra is identified with Dha */
   map['N'][0] = i++;

   /* ta-barga */
   map['t'][0] = i++; map['z'][0] = i++;
   map['d'][0] = i++; map['x'][0] = i++;
   map['n'][0] = i++;
   if (style == OLD_SORT_STYLE) map['B'][0] = map['t'][0]; /* khanda-ta is identified with ta */

   /* pa-barga */
   map['p'][0] = i++; map['f'][0] = i++;
   map['b'][0] = i++; map['v'][0] = i++;
   map['m'][0] = i++;

   /* antashtha barno */
   map['Y'][0] = i++;
   if (style == OLD_SORT_STYLE) map['y'][0] = i++; /* antashtha-ya is identified with antashtha-ja */
   map['r'][0] = i++; map['l'][0] = i++;
   /* antashtha ba is treated the same way as bargya ba */

   /* ushma barno */
   map['S'][0] = i++; map['F'][0] = i++;
   map['s'][0] = i++; map['h'][0] = i++;

   /* Bengali extras */
   if (style == DEFAULT_SORT_STYLE) {
      map['R'][0] = i++;   /* Da-e shunya ra */
      map['V'][0] = i++;   /* Dha-e shunya ra */
      map['y'][0] = i++;   /* antashtha-ya */
      map['B'][0] = i++;   /* khanda-ta */
      map['M'][0] = i++;   /* anuswar */
      map['H'][0] = i++;   /* bisarga */
      map['w'][0] = i++;   /* chandrabindu */
   }
}

void createJukMap (int style) {

   /* This procedure is independent of style */

   map[14][0] = map['Y'][0];  /* ja-fala */
   map[15][0] = map['r'][0];  /* ra-fala */
   map[94][0] = map['r'][0];  /* ref */

   map[25][0] = map['r'][0]; map[25][1] = map['u'][0];  /* ra + hraswa u ligature */
   map[26][0] = map['r'][0]; map[26][1] = map['U'][0];  /* ra + dirgha u ligature */
   map[27][0] = map['h'][0]; map[27][1] = map['u'][0];  /* ha + hraswa u ligature */
   map[28][0] = map['h'][0]; map[28][1] = map['W'][0];  /* ha + ri ligature */
   map[29][0] = map['g'][0]; map[29][1] = map['u'][0];  /* ga + u ligature */
   map[30][0] = map['S'][0]; map[30][1] = map['u'][0];  /* sha + u ligature */

   map[16][0] = map['s'][0]; map[16][1] = map['z'][0];
   map[17][0] = map['s'][0]; map[17][1] = map['n'][0];
   map[18][0] = map['s'][0]; map[18][1] = map['p'][0];
   map[19][0] = map['s'][0]; map[19][1] = map['p'][0]; map[19][2] = map['l'][0];
   map[20][0] = map['s'][0]; map[20][1] = map['f'][0];
   map[21][0] = map['s'][0]; map[21][1] = map['b'][0];
   map[22][0] = map['s'][0]; map[22][1] = map['m'][0];
   map[23][0] = map['s'][0]; map[23][1] = map['l'][0];
   map[24][0] = map['h'][0]; map[24][1] = map['N'][0];
   map[31][0] = map['h'][0]; map[31][1] = map['n'][0];
   map[38][0] = map['z'][0]; map[38][1] = map['b'][0];
   map[62][0] = map['h'][0]; map[62][1] = map['b'][0];
   map[64][0] = map['n'][0]; map[64][1] = map['d'][0]; map[64][2] = map['b'][0];
   map[76][0] = map['h'][0]; map[76][1] = map['m'][0];
   map[80][0] = map['Q'][0]; map[80][1] = map['c'][0];
   map[125][0] = map['h'][0]; map[125][1] = map['l'][0];
   map[127][0] = map['R'][0]; map[127][1] = map['g'][0];
   map[128][0] = map['k'][0]; map[128][1] = map['k'][0];
   map[129][0] = map['k'][0]; map[129][1] = map['T'][0];
   map[130][0] = map['k'][0]; map[130][1] = map['t'][0];
   map[131][0] = map['k'][0]; map[131][1] = map['b'][0];
   map[132][0] = map['k'][0]; map[132][1] = map['m'][0];
   map[133][0] = map['k'][0]; map[133][1] = map['r'][0];
   map[134][0] = map['k'][0]; map[134][1] = map['l'][0];
   map[135][0] = map['k'][0]; map[135][1] = map['F'][0];
   map[136][0] = map['k'][0]; map[136][1] = map['F'][0]; map[136][2] = map['N'][0];
   map[137][0] = map['k'][0]; map[137][1] = map['F'][0]; map[137][2] = map['m'][0];
   map[138][0] = map['k'][0]; map[138][1] = map['s'][0];
   map[139][0] = map['g'][0]; map[139][1] = map['x'][0];
   map[140][0] = map['g'][0]; map[140][1] = map['n'][0];
   map[141][0] = map['g'][0]; map[141][1] = map['b'][0];
   map[142][0] = map['g'][0]; map[142][1] = map['m'][0];
   map[143][0] = map['g'][0]; map[143][1] = map['l'][0];
   map[144][0] = map['G'][0]; map[144][1] = map['n'][0];
   map[145][0] = map['q'][0]; map[145][1] = map['k'][0];
   map[146][0] = map['q'][0]; map[146][1] = map['k'][0]; map[146][2] = map['F'][0];
   map[147][0] = map['q'][0]; map[147][1] = map['K'][0];
   map[148][0] = map['q'][0]; map[148][1] = map['g'][0];
   map[149][0] = map['q'][0]; map[149][1] = map['G'][0];
   map[150][0] = map['q'][0]; map[150][1] = map['m'][0];
   map[151][0] = map['c'][0]; map[151][1] = map['c'][0];
   map[152][0] = map['c'][0]; map[152][1] = map['C'][0];
   map[153][0] = map['c'][0]; map[153][1] = map['C'][0]; map[153][2] = map['b'][0];
   map[154][0] = map['c'][0]; map[154][1] = map['Q'][0];
   map[155][0] = map['j'][0]; map[155][1] = map['j'][0];
   map[156][0] = map['j'][0]; map[156][1] = map['j'][0]; map[156][2] = map['b'][0];
   map[157][0] = map['j'][0]; map[157][1] = map['J'][0];
   map[158][0] = map['j'][0]; map[158][1] = map['Q'][0];
   map[159][0] = map['j'][0]; map[159][1] = map['b'][0];
   map[161][0] = map['Q'][0]; map[161][1] = map['C'][0];
   map[162][0] = map['Q'][0]; map[162][1] = map['j'][0];
   map[163][0] = map['Q'][0]; map[163][1] = map['J'][0];
   map[164][0] = map['T'][0]; map[164][1] = map['T'][0];
   map[165][0] = map['T'][0]; map[165][1] = map['b'][0];
   map[166][0] = map['D'][0]; map[166][1] = map['D'][0];
   map[167][0] = map['N'][0]; map[167][1] = map['T'][0];
   map[168][0] = map['N'][0]; map[168][1] = map['Z'][0];
   map[169][0] = map['N'][0]; map[169][1] = map['D'][0];
   map[170][0] = map['N'][0]; map[170][1] = map['N'][0];
   map[171][0] = map['N'][0]; map[171][1] = map['m'][0];
   map[172][0] = map['t'][0]; map[172][1] = map['t'][0];
   map[173][0] = map['t'][0]; map[173][1] = map['t'][0]; map[173][2] = map['b'][0];
   map[174][0] = map['t'][0]; map[174][1] = map['z'][0];
   map[175][0] = map['t'][0]; map[175][1] = map['n'][0];
   map[176][0] = map['t'][0]; map[176][1] = map['b'][0];
   map[177][0] = map['t'][0]; map[177][1] = map['m'][0];
   map[178][0] = map['t'][0]; map[178][1] = map['r'][0];
   map[179][0] = map['t'][0]; map[179][1] = map['r'][0]; map[179][2] = map['u'][0];
   map[180][0] = map['d'][0]; map[180][1] = map['g'][0];
   map[181][0] = map['d'][0]; map[181][1] = map['G'][0];
   map[182][0] = map['d'][0]; map[182][1] = map['d'][0];
   map[183][0] = map['d'][0]; map[183][1] = map['x'][0];
   map[184][0] = map['d'][0]; map[184][1] = map['b'][0];
   map[185][0] = map['d'][0]; map[185][1] = map['v'][0];
   map[186][0] = map['d'][0]; map[186][1] = map['v'][0]; map[186][2] = map['r'][0];
   map[187][0] = map['d'][0]; map[187][1] = map['m'][0];
   map[188][0] = map['x'][0]; map[188][1] = map['n'][0];
   map[189][0] = map['x'][0]; map[189][1] = map['b'][0];
   map[190][0] = map['n'][0]; map[190][1] = map['T'][0];
   map[191][0] = map['n'][0]; map[191][1] = map['D'][0];
   map[192][0] = map['n'][0]; map[192][1] = map['t'][0];
   map[193][0] = map['n'][0]; map[193][1] = map['t'][0]; map[193][2] = map['b'][0];
   map[194][0] = map['n'][0]; map[194][1] = map['t'][0]; map[194][2] = map['b'][0];
   map[195][0] = map['n'][0]; map[195][1] = map['t'][0]; map[195][2] = map['r'][0];
   map[196][0] = map['n'][0]; map[196][1] = map['z'][0];
   map[197][0] = map['n'][0]; map[197][1] = map['d'][0];
   map[198][0] = map['n'][0]; map[198][1] = map['x'][0];
   map[199][0] = map['n'][0]; map[199][1] = map['n'][0];
   map[200][0] = map['n'][0]; map[200][1] = map['b'][0];
   map[201][0] = map['n'][0]; map[201][1] = map['m'][0];
   map[202][0] = map['n'][0]; map[202][1] = map['s'][0];
   map[203][0] = map['p'][0]; map[203][1] = map['T'][0];
   map[204][0] = map['p'][0]; map[204][1] = map['t'][0];
   map[205][0] = map['p'][0]; map[205][1] = map['n'][0];
   map[206][0] = map['p'][0]; map[206][1] = map['p'][0];
   map[207][0] = map['p'][0]; map[207][1] = map['l'][0];
   map[208][0] = map['p'][0]; map[208][1] = map['s'][0];
   map[209][0] = map['f'][0]; map[209][1] = map['l'][0];
   map[210][0] = map['b'][0]; map[210][1] = map['j'][0];
   map[211][0] = map['b'][0]; map[211][1] = map['d'][0];
   map[212][0] = map['b'][0]; map[212][1] = map['x'][0];
   map[213][0] = map['b'][0]; map[213][1] = map['b'][0];
   map[214][0] = map['b'][0]; map[214][1] = map['l'][0];
   map[215][0] = map['v'][0]; map[215][1] = map['r'][0];
   map[216][0] = map['v'][0]; map[216][1] = map['l'][0];
   map[217][0] = map['m'][0]; map[217][1] = map['n'][0];
   map[218][0] = map['m'][0]; map[218][1] = map['p'][0];
   map[219][0] = map['m'][0]; map[219][1] = map['f'][0];
   map[220][0] = map['m'][0]; map[220][1] = map['b'][0];
   map[221][0] = map['m'][0]; map[221][1] = map['v'][0];
   map[222][0] = map['m'][0]; map[222][1] = map['v'][0]; map[222][2] = map['r'][0];
   map[223][0] = map['m'][0]; map[223][1] = map['m'][0];
   map[224][0] = map['m'][0]; map[224][1] = map['l'][0];
   map[225][0] = map['l'][0]; map[225][1] = map['k'][0];
   map[226][0] = map['l'][0]; map[226][1] = map['g'][0];
   map[227][0] = map['l'][0]; map[227][1] = map['g'][0]; map[227][2] = map['u'][0];
   map[228][0] = map['l'][0]; map[228][1] = map['T'][0];
   map[229][0] = map['l'][0]; map[229][1] = map['D'][0];
   map[230][0] = map['l'][0]; map[230][1] = map['p'][0];
   map[231][0] = map['l'][0]; map[231][1] = map['b'][0];
   map[232][0] = map['l'][0]; map[232][1] = map['m'][0];
   map[233][0] = map['l'][0]; map[233][1] = map['l'][0];
   map[234][0] = map['S'][0]; map[234][1] = map['c'][0];
   map[235][0] = map['S'][0]; map[235][1] = map['C'][0];
   map[236][0] = map['S'][0]; map[236][1] = map['n'][0];
   map[237][0] = map['S'][0]; map[237][1] = map['m'][0];
   map[238][0] = map['S'][0]; map[238][1] = map['l'][0];
   map[239][0] = map['S'][0]; map[239][1] = map['b'][0];
   map[240][0] = map['F'][0]; map[240][1] = map['k'][0];
   map[241][0] = map['F'][0]; map[241][1] = map['k'][0]; map[241][2] = map['r'][0];
   map[242][0] = map['F'][0]; map[242][1] = map['T'][0];
   map[243][0] = map['F'][0]; map[243][1] = map['Z'][0];
   map[244][0] = map['F'][0]; map[244][1] = map['N'][0];
   map[245][0] = map['F'][0]; map[245][1] = map['p'][0];
   map[246][0] = map['F'][0]; map[246][1] = map['f'][0];
   map[247][0] = map['F'][0]; map[247][1] = map['m'][0];
   map[248][0] = map['s'][0]; map[248][1] = map['k'][0];
   map[249][0] = map['s'][0]; map[249][1] = map['k'][0]; map[249][2] = map['r'][0];
   map[250][0] = map['s'][0]; map[250][1] = map['k'][0]; map[250][2] = map['l'][0];
   map[251][0] = map['s'][0]; map[251][1] = map['K'][0];
   map[252][0] = map['s'][0]; map[252][1] = map['T'][0];
   map[253][0] = map['s'][0]; map[253][1] = map['t'][0];
   map[254][0] = map['s'][0]; map[254][1] = map['t'][0]; map[254][2] = map['u'][0];
   map[255][0] = map['s'][0]; map[255][1] = map['t'][0]; map[255][2] = map['r'][0];
}

void createPuncMap (int style) {

   /* This procedure is independent of style */

   map[123][0] = map[123][1] = 45;
   map[124][0] = map[124][1] = map[124][2] = 45;
   map[60][0] = map[60][1] = 46;
   map[92][0] = map[92][1] = 96;
   map[34][0] = map[34][1] = 39;
}

void trim (char *line) {
   int i, j, len;
   char *modline = NULL, spaceadded;

   if (!(line && *line)) return;

   len = strlen(line);
   modline = (char *)malloc((len + 1) * sizeof(char));
   strcpy(modline, line);
   i = 0;
   while (((modline[i] == ' ') || (modline[i] == '\t')) && (i < len)) i++;
   for (j=i; j<=len; j++) modline[j-i] = modline[j];

   len = strlen(modline);
   if (len > 0) {
      i = len - 1;
      while ((modline[i] == ' ') || (modline[i] == '\t')) modline[i--] = 0;
   }

   len = strlen(modline);
   spaceadded = 0;
   for (i=0, j=0; i<=len; i++) {
      if ((modline[i] != ' ') && (modline[i] != '\t')) {
         line[j++] = modline[i]; spaceadded = 0;
      } else if (!spaceadded) {
         if (modline[i-1] != '\\') {
            line[j++] = ' '; spaceadded = 1;
         } else line[j++] = ' ';
      }
   }

   free(modline);
}

void process (char *line) {
   char *cmd = NULL, **arg = NULL;
   int i, j, k, narg = 0;

   if (line[0] == '!') {
      char *syscmd = NULL;

      syscmd = (char *)malloc((strlen(line) + 10) * sizeof(char));
      sprintf(syscmd, "sh -c \"%s\"", &line[1]);
      system(syscmd);
      free(syscmd); syscmd = NULL;
      return;
   }

   j = 0; while ((line[j] != ' ') && (line[j] != '\t') && (line[j] != 0)) j++;
   cmd = (char *)malloc((j + 1) * sizeof(char));
   cmd[j] = 0; for (k=j-1; k>=0; k--) cmd[k] = line[k];
   while (line[j]) {
      i = ++j;
      while (((line[j] != ' ') && (line[j] != '\t') && (line[j] != 0)) ||
             ((line[j] == ' ') && (line[j-1] == '\\'))) j++;
      arg = (char **)realloc(arg, (narg + 1) * sizeof(char *));
      arg[narg] = NULL;
      arg[narg] = (char *)malloc((j - i + 1) * sizeof(char));
      arg[narg][j-i] = 0;
      for (k=j-1; k>=i; k--) arg[narg][k-i] = line[k];
      narg++;
   }

   if ( (!strcmp(cmd, "exit")) || (!strcmp(cmd, "quit")) ){
      exitBwsort(stdout);
   } else if (!strcmp(cmd, "version")) {
      printVer(stdout);
   } else if (!strcmp(cmd, "help")) {
      if (narg == 0) printHelp(stdout, "default");
      else printHelp(stdout, arg[0]);
   } else if (!strcmp(cmd, "comp")) {
      if (narg < 2) fprintf(stderr, "comp: Too few arguments\n");
      else if (narg > 2) fprintf(stderr, "comp: Too many arguments\n");
      else {
         printf("%d\n", compare(arg[0], arg[1], 0));
      }
   } else if (!strcmp(cmd, "load")) {
      if (narg == 0) fprintf(stderr, "load: No file(s) specified\n");
      else {
         loadFile(arg[0]);
         for (i=1; i<narg; i++) addFile(arg[i]);
      }
   } else if (!strcmp(cmd, "add")) {
      if (narg == 0) fprintf(stderr, "add: No file(s) specified\n");
      else {
         for (i=0; i<narg; i++) addFile(arg[i]);
      }
   } else if (!strcmp(cmd, "show")) {
      showList();
   } else if (!strcmp(cmd, "save")) {
      if (narg == 0) fprintf(stderr, "save: No file(s) specified\n");
      else if (narg > 1) fprintf(stderr, "save: Too many arguments\n");
      else {
         saveFile(arg[0]);
      }
   } else if (!strcmp(cmd, "sortstyle")) {
      if (narg == 0) fprintf(stdout, "The current sort style is %s.\n", (sortstyle == DEFAULT_SORT_STYLE) ? "DEFAULT" : "OLD");
      else if (!strcmp(arg[0], "DEFAULT")) createMap(sortstyle = DEFAULT_SORT_STYLE);
      else if (!strcmp(arg[0], "OLD")) createMap(sortstyle = OLD_SORT_STYLE);
      else fprintf(stderr, "Unknown sort style : %s\n", arg[0]);
   } else {
      fprintf(stderr,"%s: Unknown command. Try \"help\".\n", cmd);
   }

   free(cmd);
   for (i=0; i<narg; i++) { free(arg[i]); arg[i] = NULL; }
   if (arg != NULL) free (arg);
}

void exitBwsort ( FILE *fp ) {
   fprintf(fp, "Have a nice day!\n");
   exit(0);
}

void printVer ( FILE *fp ) {
   fprintf(fp, "bwsort vesrion 1.0 (December 1998)\nAuthor: Abhijit Das (Barda) [abhij@csa.iisc.ernet.in]\n");
}

void printHelp ( FILE *fp, char topic[] ) {
   if (!(strcmp(topic, "default"))) {
      fprintf(fp, "Usage: bwsort [-s scheme] [-h] [-v] [file1 [file2 ... ]]\n");
      fprintf(fp, "If arguments are specified in the command line,  bwsort concatenates\n");
      fprintf(fp, "the files, sorts the concatenated list  and  outputs the sorted list\n");
      fprintf(fp, "to stdout. The options are\n");
      fprintf(fp, "     -s scheme   Set scheme to DEFAULT or OLD\n");
      fprintf(fp, "     -h          Print this help message\n");
      fprintf(fp, "     -v          Print version info\n");
      fprintf(fp, "The value of the  environment variable  BWSORTSTYLE, if set, is used\n");
      fprintf(fp, "as the default sorting style. This can be overridden by the -s flag.\n");
      fprintf(fp, "If no file names are specified in  the command line,  bwsort runs in\n");
      fprintf(fp, "interactive mode. In this mode, the following commands can be used:\n");
      fprintf(fp, "     add file1 [file2 [file3 ...]]\n");
      fprintf(fp, "     comp word1 word2\n");
      fprintf(fp, "     exit / quit\n");
      fprintf(fp, "     help [topic]\n");
      fprintf(fp, "     load file1 [file2 [file3 ...]]\n");
      fprintf(fp, "     save file\n");
      fprintf(fp, "     show\n");
      fprintf(fp, "     sortstyle [style]\n");
      fprintf(fp, "     version\n");
      fprintf(fp, "     !shell command\n");
      fprintf(fp, "Type \"help cmd\" at bwsort prompt to know more about the command \"cmd\".\n");
   } else if (!(strcmp(topic, "add"))) {
      fprintf(fp, "Usage: add file1 [file2 [file3 ...]]\n");
      fprintf(fp, "Add file1, file2, file3, ... to the current sorted list.\n");
   } else if (!(strcmp(topic, "comp"))) {
      fprintf(fp, "Usage: comp word1 word2\n");
      fprintf(fp, "Compare word1 and word2 according to Bengali alphabet.  Returns 0, 1\n");
      fprintf(fp, "and -1 according as word1 matches, follows  or  precedes word2 under\n");
      fprintf(fp, "the Bengali sorting rules.\n");
   } else if (!(strcmp(topic, "exit"))) {
      fprintf(fp, "Usage: exit\nQuit the interactive session of bwsort.\n");
   } else if (!(strcmp(topic, "help"))) {
      fprintf(fp, "Usage: help [topic]\n");
      fprintf(fp, "Print help message. The topic is either empty or one of the commands\n");
      fprintf(fp, "that can be interpreted during the interactive session.\n");
   } else if (!(strcmp(topic, "load"))) {
      fprintf(fp, "Usage: load file1 [file2 [file3 ...]]\n");
      fprintf(fp, "Clear the current sorted list and load file1, file2, file3 ...\n");
   } else if (!(strcmp(topic, "quit"))) {
      fprintf(fp, "Usage: quit\nQuit the interactive session of bwsort.\n");
   } else if (!(strcmp(topic, "save"))) {
      fprintf(fp, "Usage: save file\nSave the current sorted list in file.\n");
   } else if (!(strcmp(topic, "show"))) {
      fprintf(fp, "Usage: show\nList the current sorted list on screen.\n");
   } else if (!(strcmp(topic, "sortstyle"))) {
      fprintf(fp, "Usage: sortstyle [style]\n");
      fprintf(fp, "Print or set the sorting style. The style is either DEFAULT or OLD. If\n");
      fprintf(fp, "the environment variable BWSORTSTYLE is set, it's value is used as the\n");
      fprintf(fp, "default. Otherwise, the default style is DEFAULT.\n");
   } else if (!(strcmp(topic, "version"))) {
      fprintf(fp, "Usage: version\nPrint version info.\n");
   } else if (!(strcmp(topic, "!"))) {
      fprintf(fp, "Usage: !cmd [arg1 [arg2 ...]]\n");
      fprintf(fp, "Execute the shell command \"cmd\" with arguments arg1, arg2 ...\n");
   } else fprintf(fp, "%s: Unknown topic\n", topic);
}

void convert (unsigned char inp[], unsigned char *outp) {
   int i, j, k, len, done;
   unsigned char kaar, banjon, hasanta, jafala, rafala, ref, chandra;

   len = strlen(inp);

   for (i = 0, j = 0; i<len; i++) {
      if ( (map[inp[i]][0] <= 64) || (map[inp[i]][0] >= 123) ) {
         k = 0; while (outp[j++] = map[inp[i]][k++]); j--;
      } else if ( (map[inp[i]][0] >= 91) && (map[inp[i]][0] <= 96) ) {
         k = 0; while (outp[j++] = map[inp[i]][k++]); j--;
      } else if (inp[i] == 'a') {
         done = chandra = kaar = jafala = 0; i++;
         while ( (!done) && (i < len) ) {
            switch (inp[i]) {
               case 'A': kaar = 1; break;
               case 14: jafala = 1; break;
               case 'w': chandra = 1; break;
               default: done = 1;
            }
            if (!done) i++;
         }
         i--;
         if (kaar && jafala) outp[j++] = map[9][0];
         else if (kaar) outp[j++] = map['A'][0];
         else outp[j++] = map['a'][0];
         if (chandra) outp[j++] = map['w'][0];
      } else if (inp[i] == 6) {
         done = chandra = kaar = jafala = 0; i++;
         while ( (!done) && (i < len) ) {
            switch (inp[i]) {
               case 'A': kaar = 1; break;
               case 14: jafala = 1; break;
               case 'w': chandra = 1; break;
               default: done = 1;
            }
            if (!done) i++;
         }
         i--;
         if (kaar && jafala) outp[j++] = map[10][0];
         else outp[j++] = map['e'][0];
         if (chandra) outp[j++] = map['w'][0];
      } else if ( ((inp[i] >= 1) && (inp[i] <= 8)) || (inp[i] == 'o') ) {
         outp[j++] = map[inp[i]][0];
      } else {
         if ((inp[i] == 'i') || (inp[i] == 'e') || (inp[i] == 'E'))
            kaar = inp[i++];
         else kaar = 'a';
         banjon = inp[i++];
         switch (banjon) {
            case 25: banjon = 'r'; kaar = 'u'; break;
            case 26: banjon = 'r'; kaar = 'U'; break;
            case 27: banjon = 'h'; kaar = 'u'; break;
            case 28: banjon = 'h'; kaar = 'W'; break;
            case 29: banjon = 'g'; kaar = 'u'; break;
            case 30: banjon = 'S'; kaar = 'u'; break;
            case 179: banjon = 178; kaar = 'u'; break;
            case 193: banjon = 192; kaar = 'u'; break;
            case 227: banjon = 226; kaar = 'u'; break;
            case 254: banjon = 253; kaar = 'u'; break;
         }
         done = 0;
         jafala = rafala = ref = chandra = hasanta = 0;
         while ((!done) && (i < len)) {
            switch (inp[i]) {
               case 'A': if (kaar == 'e') kaar = 8; else kaar = 'A'; break;
               case 'I': 
               case 'u': 
               case 'U': 
               case 'W': 
               case 'O': kaar = inp[i]; break;
               case '_': hasanta = 1; break;
               case '^': ref = 1; break;
               case 14 : jafala = 1; break;
               case 15 : rafala = 1; break;
               case 'w': chandra = 1; break;
               default : done = 1;
            }
            if (!done) i++;
         }
         i--;
         if (ref) outp[j++] = map['r'][0];
         k = 0; while (outp[j++] = map[banjon][k++]); j--;
         if (rafala) outp[j++] = map['r'][0];
         if (jafala) outp[j++] = map['Y'][0];
         if (sortstyle == OLD_SORT_STYLE) {
            outp[j++] = map[kaar][0];
         } else if ((!hasanta) || (kaar != 'a')) {
            outp[j++] = map[kaar][0];
         }
         if (chandra) outp[j++] = map['w'][0];
      }
   }
   outp[j] = 0;
}

int compare (unsigned char one[], unsigned char two[], int flag) {
   int len1, len2, res;
   unsigned char *ONE = NULL, *TWO = NULL;

   if ( (one == NULL) || (two == NULL) ) {
      fprintf(stderr, "compare: NULL pointer(s) passed\n");
   }

   len1 = strlen(one); len2 = strlen(two);
   if (len1 == 0) {
      if (len2 == 0) return (0);
      return (-1);
   }
   if (len2 == 0) return (1);

   if (!(flag & 1)) {
      ONE = (unsigned char *)malloc((3 * len1 + 1) * sizeof(unsigned char));
      convert(one, ONE);
   } else ONE = one;
   if (!(flag & 2)) {
      TWO = (unsigned char *)malloc((3 * len2 + 1) * sizeof(unsigned char));
      convert(two, TWO);
   } else TWO = two;

   res = strcmp(ONE, TWO);
   if (!(flag & 1)) free(ONE); if (!(flag & 2)) free(TWO);
   if (res == 0) return (0);
   if (res < 0) return (-1);
   return (1);
}

void insert (unsigned char *nextline, unsigned char *convline) {
   int i, j, inserted, compres;

   if (convline == NULL) return;
   if (strlen(convline) == 0) return;

   inserted = i = 0;
   while ( (!inserted) && (i < nword) ) {
      compres = strcmp(convline, clist[i]);
      if (compres <= 0) {
         if (nword == assigned) {
            assigned += 100;
            list = (unsigned char **)realloc(list, assigned * sizeof(unsigned char *));
            clist = (unsigned char **)realloc(clist, assigned * sizeof(unsigned char *));
            for (j = nword; j < assigned; j++) list[j] = clist[j] = NULL;
         }
         for (j = nword; j > i; j--) {
            list[j] = list[j-1]; list[j-1] = NULL;
            clist[j] = clist[j-1]; clist[j-1] = NULL;
         }
         list[i] = (unsigned char *)malloc((strlen(nextline) + 1) * sizeof(unsigned char));
         strcpy(list[i], nextline);
         free(clist[i]); clist[i] = NULL;
         clist[i] = (unsigned char *)malloc((strlen(convline) + 1) * sizeof(unsigned char));
         strcpy(clist[i], convline);
         nword++;
         inserted = 1;
      } else i++;
   }
   if (!inserted) {
      if (nword == assigned) {
         assigned += 100;
         list = (unsigned char **)realloc(list, assigned * sizeof(unsigned char *));
         clist = (unsigned char **)realloc(clist, assigned * sizeof(unsigned char *));
         for (j=nword; j<assigned; j++) list[j] = clist[j] = NULL;
      }
      list[nword] = (unsigned char *)malloc((strlen(nextline) + 1) * sizeof(unsigned char));
      strcpy(list[nword], nextline);
      clist[nword] = (unsigned char *)malloc((strlen(convline) + 1) * sizeof(unsigned char));
      strcpy(clist[nword], convline);
      nword++;
   }
}

void loadFile ( char fname[] ) {
   int i;
   FILE *fp;
   unsigned char ch, nextline[1024], convline[3072];

   if (assigned > 0) {
      if (list != NULL) {
         for (i=0; i<assigned; i++) {
            if (list[i] != NULL) { free(list[i]); list[i] = NULL; }
         }
         free(list); list = NULL;
      }
      if (clist != NULL) {
         for (i=0; i<assigned; i++) {
            if (clist[i] != NULL) { free(clist[i]); clist[i] = NULL; }
         }
         free(clist); clist = NULL;
      }
   }
   nword = assigned = 0;

   fp = (FILE *)fopen(fname, "r");
   if (fp == NULL) {
      fprintf(stderr, "Unable to read file %s\n", fname);
      return;
   }

   fprintf(stderr, "Reading file %s ..", fname); fflush(stdout);
   while (!feof(fp)) {
      i = 0;
      do {
         fscanf(fp, "%c", &ch);
         if (ch != '\n') {
            if (i == 1023) {
               fprintf(stderr, "Line too long in file %s\n", fname);
               fclose(fp);
               return;
            }
            nextline[i++] = ch;
         }
      } while ((ch != '\n') && (!feof(fp)));
      nextline[i] = 0;
      convert(nextline, convline);
      insert(nextline, convline);
      if (!(nword & 255)) { fprintf(stderr, "."); fflush(stderr); }
   }
   fprintf(stderr, ". Done\n");

   fclose(fp);
}

void addFile ( char fname[] ) {
   int i;
   FILE *fp;
   unsigned char ch, nextline[1024], convline[3072];

   fp = (FILE *)fopen(fname, "r");
   if (fp == NULL) {
      fprintf(stderr, "Unable to read file %s\n", fname);
      return;
   }

   fprintf(stderr, "Reading file %s ..", fname); fflush(stderr);
   while (!feof(fp)) {
      i = 0;
      do {
         fscanf(fp, "%c", &ch);
         if (ch != '\n') {
            if (i == 1023) {
               fprintf(stderr, "Line too long in file %s\n", fname);
               fclose(fp);
               return;
            }
            nextline[i++] = ch;
         }
      } while ((ch != '\n') && (!feof(fp)));
      nextline[i] = 0;
      convert(nextline, convline);
      insert(nextline, convline);
      if (!(nword & 255)) { fprintf(stderr, "."); fflush(stderr); }
   }
   fprintf(stderr, ". Done\n");

   fclose(fp);
}

void showList () {
   int i;

   for (i=0; i<nword; i++) printf("%s\n",list[i]);
   printf("Total %d lines\n", nword);
}

void saveFile ( char fname[] ) {
   int i;
   FILE *fp;

   fp = (FILE *)fopen(fname, "w");
   if (fp == NULL) {
      fprintf(stderr, "Unable to write file %s\n", fname);
      return;
   }

   for (i=0; i<nword; i++) fprintf(fp, "%s\n", list[i]);
   fclose(fp);
   printf("File %s saved\n", fname);
}

int main (int argc, int *argv[]) {

   initialize();

   if (argc > 1) {
      int i, nfile;
      char **fname = NULL;

      i = 1; nfile = 0;
      while (i < argc) {
         trim((char *)argv[i]);
         if (!strcmp((char *)argv[i], "-h")) { printHelp(stderr, "default"); exit(0); }
         if (!strcmp((char *)argv[i], "-v")) { printVer(stderr); exit(0); }
         if (!strcmp((char *)argv[i], "-s")) {
            if (i == argc - 1) {
               fprintf(stderr, "No argument for option -s\n");
               exit(1);
            } else if (!strcmp((char *)argv[++i], "DEFAULT")) sortstyle = DEFAULT_SORT_STYLE;
            else if (!(char *)strcmp((char *)argv[i], "OLD")) sortstyle = OLD_SORT_STYLE;
            else {
               fprintf(stderr, "Unknown sorting style %s\n", argv[i]);
               exit(1);
            }
         } else if (strlen((char *)argv[i]) > 0) {
            fname = (char **)realloc(fname, (nfile + 1) * sizeof(char *));
            fname[nfile] = NULL;
            fname[nfile] = (char *)malloc((strlen((char *)argv[i]) + 1) * sizeof(char));
            strcpy(fname[nfile], (char *)argv[i]); nfile++;
         }
         i++;
      }
      if (nfile > 0) {
         createMap(sortstyle);
         for (i=0; i<nfile; i++) addFile(fname[i]);
         for (i=0; i<nword; i++) fprintf(stdout, "%s\n", list[i]);
         exit(0);
      }
   }

   createMap(sortstyle);

   while (1) {
      if (line != NULL) { free(line); line = NULL; }
      line = readline("bwsort> ");
      trim(line);
      if (line && *line) {
         add_history (line);
         process(line);
      }
   }

   exit(0);
}

